#include "metacraneclient.h"
#include <QJsonObject>
#include <QJsonDocument>
#include <QDesktopServices>
#include <QUrl>
#include <QTimer>
#include <QByteArray>
#include <QFile>
#include <QFileInfo>

const size_t BUFFER_SIZE = 250000;

MetacraneClient::MetacraneClient(QObject *parent) :
    QObject(parent)
,   m_next_transfer_id(1)
,   m_current_file(NULL)
,   m_current_file_stream(NULL)
,   m_current_transfer_id(-1)
,   m_current_transfer_num_pieces(0)
,   m_current_transfer_piece_index(0)
{
    QObject::connect(&m_socket, SIGNAL(connected()), this, SLOT(socketConnected()));
    QObject::connect(&m_socket, SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
    addBuiltinHandlers();
    m_socket.pokeConnection();
    QTimer *timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(update()));
    timer->start(5000);
}

MetacraneClient::~MetacraneClient()
{
    delete m_current_file_stream;
    delete m_current_file;
}

void MetacraneClient::pokeConnection()
{
    m_socket.pokeConnection();
}

void MetacraneClient::gotoCrane(QString name)
{
    if(!m_socket.isConnected()){
        qDebug()<<"gotoCrane:"<<name<<": delaying until connected";
        pokeConnection();
        m_current_crane_name = name;
        currentCraneNameChanged();
        return;
    }
    qDebug()<<"gotoCrane:"<<name;
    QVariantMap params;
    params["name"] = name;
    sendRequest("join_crane", params,
            [this, name](const QVariantMap &response)
    {
        qDebug()<<"join_crane response"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            return;
        }
        m_current_crane_name = name;
        currentCraneNameChanged();
    });
}

void MetacraneClient::openUrl(QString url)
{
    QDesktopServices::openUrl(QUrl(url));
}

void MetacraneClient::clearMessage()
{
    m_last_message.clear();
    lastMessageChanged();
}

void MetacraneClient::uploadText(QString text)
{
    QVariantMap params;
    params["text"] = text;
    params["confirm_downloads"] = false;
    params["deletion_time"] = 56 * 3600;
    sendRequest("lift_text", params,
            [this](const QVariantMap &response)
    {
        qDebug()<<"lift_text response"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            return;
        }
    });
}

void MetacraneClient::uploadFile(QString path)
{
    delete m_current_file_stream;
    m_current_file_stream = NULL;
    delete m_current_file;
    m_current_file = NULL;

    m_current_file = new QFile(path);
    QFile &f = *m_current_file;
    if(!f.open(QIODevice::ReadOnly)){
        showMessage("Error: Could not open file for reading");
        return;
    }

    QFileInfo info(f);
    quint64 size = info.size();
    qDebug()<<"upload file size:"<<size;

    m_current_file_stream = new QDataStream(m_current_file);
    m_current_transfer_id = m_next_transfer_id++;
    m_current_transfer_num_pieces = ((size - 1) / BUFFER_SIZE) + 1;
    m_current_transfer_piece_index = 0;

    QVariantMap params;
    params["transfer_id"] = m_current_transfer_id;
    params["name"] = info.fileName();
    params["num_pieces"] = m_current_transfer_num_pieces;
    params["size"] = size;
    params["confirm_downloads"] = false;
    params["deletion_time"] = 56*3600;
    sendRequest("start_file_transfer", params,
            [this](const QVariantMap &response){
        qDebug()<<"start_file_transfer response:"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            return;
        }
        sendNextFilePiece();
    });
    uploadProgressChanged();
}

void MetacraneClient::addMessageHandler(QString name,MetacraneSocket::msg_cb_t handler)
{
    m_socket.addMessageHandler(name, handler);
}

double MetacraneClient::uploadProgress()
{
    if(m_current_file_stream == NULL || m_current_file == NULL)
        return -1;
    return (double)m_current_transfer_piece_index / m_current_transfer_num_pieces;
}

void MetacraneClient::addBuiltinHandlers()
{
    addMessageHandler("message", [this](const QVariantMap &response){
        QString text = response.value("text", QString()).toString();
        qDebug()<<"message"<<text;
        m_last_message = "Server says: "+text;
        lastMessageChanged();
    });
    addMessageHandler("add_entity", [this](const QVariantMap &response){
        int entity = response.value("entity", -1).toInt();
        qDebug()<<"add_entity"<<entity;
        QVariantMap components = response.value("components", QVariant()).toMap();
        Entity *e = new Entity(entity);
        e->components() = components;
        m_entities.prepend(e);
        entitiesChanged();
    });
    addMessageHandler("delete_entity", [this](const QVariantMap &response){
        int entity = response.value("entity", -1).toInt();
        qDebug()<<"delete_entity"<<entity;
        QMutableListIterator<QObject*> i(m_entities);
        while (i.hasNext()) {
            Entity *e = (Entity*)i.next();
            if(e->entity() == entity)
                i.remove();
        }
        entitiesChanged();
    });
    // TODO: add_component
    // TODO: delete_component
}

void MetacraneClient::sendMessage(QString name, const QVariantMap &params)
{
    m_socket.sendMessage(name, params);
}

void MetacraneClient::sendRequest(QString name, const QVariantMap &params,
        std::function<void(const QVariantMap &response)> cb)
{
    m_socket.sendRequest(name, params, cb);
}

void MetacraneClient::showMessage(QString message)
{
    m_last_message = message;
    lastMessageChanged();
}

void MetacraneClient::sendNextFilePiece()
{
    if(m_current_file_stream == NULL || m_current_file == NULL)
        return;
    qDebug()<<"Sending piece"<<m_current_transfer_piece_index;
    qDebug()<<"m_file_stream"<<m_current_file_stream;
    QByteArray buf(BUFFER_SIZE, 0);
    int read_len = m_current_file_stream->readRawData(buf.data(), BUFFER_SIZE);
    qDebug()<<"read_len:"<<read_len;
    if(read_len != BUFFER_SIZE &&
            m_current_transfer_piece_index != m_current_transfer_num_pieces - 1){
        qDebug()<<"Read wrong size:"<<read_len;
        showMessage("Error: Failed to read full buffer");
        return;
    }
    QByteArray buf2(buf.data(), read_len);

    QVariantMap params;
    params["transfer_id"] = m_current_transfer_id;
    params["index"] = m_current_transfer_piece_index++;
    params["data_b64"] = buf2.toBase64();
    sendRequest("file_data", params,
            [this](const QVariantMap &response){
        qDebug()<<"file_data response:"<<response;
        if(response["success"].toBool() == false){
            qDebug()<<response["error"];
            showMessage("Error: "+response["error"].toString());
            return;
        }
        if(m_current_transfer_piece_index < m_current_transfer_num_pieces){
            sendNextFilePiece();
            return;
        }
        qDebug()<<"Transfer finished";
        delete m_current_file_stream;
        m_current_file_stream = NULL;
        delete m_current_file;
        m_current_file = NULL;
        uploadProgressChanged();
    });
    uploadProgressChanged();
}

void MetacraneClient::socketConnected()
{
    qDebug()<<"client connected";

    m_entities.clear();

    QVariantMap params;
    params["version"] = 2;
    sendMessage("version", params);

    if(m_current_crane_name != ""){
        qDebug()<<"joining delayed crane";
        gotoCrane(m_current_crane_name);
    }
}

void MetacraneClient::socketDisconnected()
{
    qDebug()<<"client disconnected";
}

void MetacraneClient::update()
{
    m_socket.pokeConnection();
}
